__all__ = ['TaskManager']
__author__ = "Nikolay Grishchenko"
__version__ = "1.0"

import boto3
import json
import logging
import os
import time
import uuid

from copy import deepcopy
from json.decoder import JSONDecodeError
from pkg_resources import parse_version
from typing import Dict, List, Optional, Union

from sosw.app import Processor
from sosw.components.benchmark import benchmark
from sosw.components.dynamo_db import DynamoDbClient
from sosw.components.helpers import first_or_none
# from sosw.managers.ecology import EcologyManager
from sosw.labourer import Labourer


logger = logging.getLogger()
logger.setLevel(logging.INFO)


class TaskManager(Processor):
    """.. _task:

    TaskManager is the core class used by most SOSW Lambdas.
    It handles all the operations with tasks thus the configuration of this Manager is essential during your SOSW implementation.

    The default version of TaskManager works with DynamoDB tables to store and analyze the state of Tasks.
    This could be upgraded in future versions to work with other persistent storage or DBs.

    The very important concept to understand about Task workflow is `greenfield`. :ref:`Read more <greenfield>`.

    """

    DEFAULT_CONFIG = {
        'init_clients':                            ['DynamoDb', 'lambda', 'Ecology'],
        'dynamo_db_config':                        {
            'table_name':       'sosw_tasks',
            'index_greenfield': 'sosw_tasks_greenfield',
            'row_mapper':       {
                'task_id':             'S',
                'labourer_id':         'S',
                'created_at':          'N',
                'completed_at':        'N',
                'greenfield':          'N',
                'attempts':            'N',
                'closed_at':           'N',
                'desired_launch_time': 'N',
                'arn':                 'S',
                'payload':             'S'
            },
            'required_fields':  ['task_id', 'labourer_id', 'created_at', 'greenfield'],

            # You can overwrite field names to match your DB schema. But the types should be the same.
            # By default takes the key itself.
            'field_names':      {
                'task_id': 'task_id',  # This is just an example
            }
        },
        'sosw_closed_tasks_table':                 'sosw_closed_tasks',
        'sosw_closed_tasks_labourer_status_index': 'labourer_task_status_with_time',
        'sosw_retry_tasks_table':                  'sosw_retry_tasks',
        'sosw_retry_tasks_greenfield_index':       'labourer_id_greenfield',
        'greenfield_invocation_delta':             31557600,  # 1 year.
        'greenfield_task_step':                    1000,
        'labourers':                               {
            # 'some_function': {
            #     'arn':                          'arn:aws:lambda:us-west-2:0000000000:function:some_function',
            #     'max_simultaneous_invocations': 10,
            #     # Health metrics for this Labourer should be stored in a dictionary.
            #     'health_metrics':               {
            #     # Name of the metric is just for human readability (probaly some future GUI interfaces),
            #         'SomeDBCPU': {
            #             # The value must have ``'details'`` as a dict with kwargs for CloudWatch client.
            #             'details':                     {
            #                 'Name':       'CPUUtilization',
            #                 'Namespace':  'AWS/RDS',
            #                 'Period':     60,
            #                 'Statistics': ['Average'],
            #                 'Dimensions': [
            #                     {
            #                         'Name':  'DBInstanceIdentifier',
            #                         'Value': 'YOUR-DB'
            #                     },
            #                 ],
            #             },
            #
            #             # These is the mapping of how the Labourer should "feel" about this metric.
            #             # See EcologyManager.ECO_STATUSES.
            #             # This is just a mapping ``ECO_STATUS: value`` using ``feeling_comparison_operator``.
            #             'feelings':                    {
            #                 3: 50,
            #                 4: 25,
            #             },
            #             'feeling_comparison_operator': '<='
            #         },
            #     },
            # },
        },
        'max_attempts':                            3,
        'max_closed_to_analyse_for_duration':      10,
        'max_simultaneous_invocations':            1,
    }

    __labourers = None

    # these clients will be initialized by Processor constructor
    # ecology_client: EcologyManager = None
    ecology_client = None
    dynamo_db_client: DynamoDbClient = None
    lambda_client: boto3.client = None


    def get_oldest_greenfield_for_labourer(self, labourer: Labourer, reverse: bool = False) -> int:
        """
        Return value of oldest greenfield in queue.
        This means the beginning of the queue if you need FIFO behaviour.
        """

        _ = self.get_db_field_name

        q = dict(
                keys={_('labourer_id'): labourer.id, _('greenfield'): str(time.time())},
                comparisons={_('greenfield'): '<='},
                max_items=1,
                index_name=self.config['dynamo_db_config']['index_greenfield']
        )
        if reverse:
            q['desc'] = True

        items = self.dynamo_db_client.get_by_query(**q)

        if items:
            first_task_in_queue = items[0]
            result = int(first_task_in_queue[_('greenfield')])

        # Logically this is 0 (aka beginning of the Epoch), but we sometimes want to put earlier than the oldest task,
        # so let us assume we begin one step ahead of The zero.
        else:
            result = 0 + int(self.config['greenfield_task_step'])

        logger.debug(f"get_oldest_greenfield_for_labourer(reverse={reverse}) returned: {result}")
        return result


    def get_newest_greenfield_for_labourer(self, labourer: Labourer) -> int:
        """
        Return value of the newest greenfield in queue. This means the end of the queue or latest added.
        """
        return self.get_oldest_greenfield_for_labourer(labourer, reverse=True)


    def get_length_of_queue_for_labourer(self, labourer: Labourer) -> int:
        """
        Approximate count of tasks still in queue for `labourer`.
        Tasks with greenfield <= now()

        :param labourer:
        :return:
        """

        _ = self.get_db_field_name

        queue_count = self.dynamo_db_client.get_by_query(
                keys={_('labourer_id'): labourer.id, _('greenfield'): str(time.time())},
                comparisons={'greenfield': '<='},
                index_name=self.config['dynamo_db_config']['index_greenfield'],
                return_count=True)

        return queue_count


    def register_labourers(self) -> List[Labourer]:
        """
        Sets timestamps, health status and other custom attributes on Labourer objects passed for registration.

        We also send a pointer to the TaskManager (aka self) to Ecology Manager.
        The latter will have to make some queries, and we don't want him to initialise another TaskManager for himself.
        """

        _ = self.get_db_field_name
        _cfg = self.config.get

        self.ecology_client.register_task_manager(self)

        # WARNING! This must be something ordered, because these methods depend on one another.
        custom_attributes = (
            ('start', lambda x: int(time.time())),
            ('invoked', lambda x: x.get_attr('start') + self.config['greenfield_invocation_delta']),
            ('expired', lambda x: x.get_attr('invoked') - (x.duration + x.cooldown)),
            ('health', lambda x: self.ecology_client.get_labourer_status(x)),
            ('health_metrics', lambda x: _cfg('labourers')[x.id].get('health_metrics')) or {},
            ('max_attempts', lambda x: self.config.get(f'max_attempts_{x.id}') or self.config['max_attempts']),
            ('max_duration', lambda x: self.ecology_client.get_max_labourer_duration(x)),
            ('average_duration', lambda x: self.ecology_client.get_labourer_average_duration(x)),
            ('max_simultaneous_invocations', lambda x: _cfg('labourers')[x.id].get('max_simultaneous_invocations')
                                                       or _cfg('max_simultaneous_invocations')),
        )

        # Reset old Labourers and reconstruct them with fresh data.
        self.__labourers = None
        labourers = self.get_labourers()

        result = []
        for labourer in labourers:
            for k, method in [x for x in custom_attributes]:
                labourer.set_custom_attribute(k, method(labourer))
                logger.debug(f"SET for {labourer}: {k} = {method(labourer)}")
            result.append(labourer)

            for attr, val in _cfg('labourers')[labourer.id].items():
                labourer.set_custom_attribute(attr, val)

        self.__labourers = result

        return result


    def get_labourers(self) -> List[Labourer]:
        """
        Return configured Labourers.
        Config of the TaskManager expects 'labourers' as a dict 'name_of_lambda': {'some_setting': 'value1'}
        """

        if not self.__labourers:
            self.__labourers = [Labourer(id=name, **settings) for name, settings in self.config['labourers'].items()]

        return self.__labourers


    def get_labourer(self, labourer_id: str) -> Labourer:
        return first_or_none(self.get_labourers(), lambda x: x.id == labourer_id)


    def get_db_field_name(self, key: str) -> str:
        """ Could be useful if you overwrite field names with your own ones (e.g. for tests). """
        return self.config['dynamo_db_config']['field_names'].get(key, key)


    def create_task(self, labourer: Labourer, strict: bool = True, **kwargs):
        """
        Schedule a new task.

        :param labourer:    Labourer object of Lambda to execute the task.
        :param bool strict: By default (True) prohibits specifying in the task (kwargs) the fields that are supposed
                            to be autogenerated. Only if they match with autogen - then pass. You can override this
                            and pass custom task properties setting strict = False
        """

        _ = self.get_db_field_name

        # Save a copy of kwargs, because we are going to play with them.
        kw = deepcopy(kwargs)

        # Some common function we may need to generate default values.
        autogenerators = {
            _('task_id'):     lambda: str(uuid.uuid1().hex),
            _('labourer_id'): lambda: str(labourer.id),
            _('created_at'):  lambda: str(time.time()),
            _('greenfield'):  lambda: str(self.get_newest_greenfield_for_labourer(labourer)
                                          + int(self.config['greenfield_task_step'])),
            _('attempts'):    lambda: '0',
        }

        new_task = {}

        # Fill task fields from either kwargs or autogenerator.
        for key, gen in autogenerators.items():
            if strict:
                suggested = kw.pop(key, None)
                if suggested:
                    if str(suggested) != gen():
                        raise ValueError(f"Value of {key} passed to `create_task` doesnot match autogenerated. "
                                         f"Use strict = False if you really have to create a task like this.")
                else:
                    suggested = gen()

            else:
                suggested = str(kw.pop(key, gen()))
            new_task[key] = suggested

        if not all(key in new_task for key in self.config['dynamo_db_config'].get('required_fields', [])):
            raise ValueError(f"Required key {key} is missing in task {kwargs} and "
                             f"we don't have any auto generator for it.")

        try:
            new_task['payload'] = self.construct_payload_for_task(**kw)
        except Exception:
            raise ValueError(f"Unexpected `payload` or custom attrs for task '{kwargs}'. Should be dict() or JSON.")

        # Saving to DynamoDB.
        self.dynamo_db_client.put(new_task)
        logger.debug(f"Created a task: {new_task}")


    def construct_payload_for_task(self, **kwargs) -> str:
        """
        Combines remaining kwargs to a singular JSON payload.
        """

        _ = self.get_db_field_name

        # Now create and fill the payload
        payload = kwargs.pop(_('payload'), dict())

        # We have to use a dictionary to add more elements soon.
        if isinstance(payload, str):
            try:
                result = json.loads(payload)
            except Exception:
                result = dict(payload=payload)
        elif not isinstance(payload, dict):
            result = dict(payload=payload)
        else:
            result = payload

        # Add other fields from kwargs to payload.
        for key, value in kwargs.items():
            result[key] = value

        logger.debug(f"Constructed payload ({result})from {kwargs}")
        return json.dumps(result)


    def is_valid_task(self, task: Dict) -> bool:
        """
        Simple validation for required fields.
        """
        _ = self.get_db_field_name
        _cfg = self.config.get

        return all(field in task.keys() for field in [_('task_id'), _('labourer_id'), _('created_at')])


    def invoke_task(self, labourer: Labourer, task_id: Optional[str] = None, task: Optional[Dict] = None):
        """
        Invoke the Lambda Function execution for `task`.
        Providing the ID is more expensive, but safer from "task injection" attacks method that prefetches Task
        from the table before trying to invoke.

        Skips already running tasks with no exception, thus concurrent Orchestrators (or whoever else)
        should not duplicate invocations.
        """

        if not any([task, task_id]) or all([task, task_id]):
            raise AttributeError(f"You must provide any of `task` or `task_id`.")

        if not task:
            task = self.get_task_by_id(task_id=task_id)

        if not self.is_valid_task(task):
            raise ValueError(f"Task to invoke is invalid: {task}")

        try:
            self.mark_task_invoked(labourer, task)
        except Exception as err:
            if err.__class__.__name__ == 'ConditionalCheckFailedException':
                logger.warning(f"Update failed due to already running task {task}. "
                               f"Probably concurrent Orchestrator already invoked.")
                self.stats['concurrent_task_invocations_skipped'] += 1
                return
            else:
                logger.exception(err)
                raise RuntimeError(err)

        # Flatten the payload
        call_payload = task.pop('payload', {})
        if isinstance(call_payload, str):
            try:
                call_payload = json.loads(call_payload)
            except JSONDecodeError as err:
                logger.exception(f"Failed to decode payload: {call_payload}. Probably invalid task.")
                # TODO should panic, but not die here. Just skip the task.

        call_payload.update(task)

        # Support for asyncio does not exist for botocore right now. See ticket below:
        # https://github.com/boto/botocore/issues/458
        # There is another library that makes an async version of botocore, but doesn't
        # support lambda currently. See below:
        # https://github.com/aio-libs/aiobotocore
        lambda_response = self.lambda_client.invoke(
                FunctionName=labourer.arn,
                InvocationType='Event',
                Payload=json.dumps(call_payload)
        )
        logger.debug(lambda_response)

        self.stats['invoked_tasks'] += 1


    def mark_task_invoked(self, labourer: Labourer, task: Dict, check_running: Optional[bool] = True):
        """
        Update the greenfield with the latest invocation timestamp + invocation_delta

        By default updates with a conditional expression that fails in case the current greenfield is already in
        `invoked` state. If this check fails the function raises RuntimeError that should be handled
        by the Orchestrator. This is very important to help duplicate invocations of the Worker by simultaneously
        running Orchestrators.

        :param labourer:        Labourer for the task
        :param task:            Task dictionary
        :param check_running:   If True (default) updates with conditional expression.

        :raises RuntimeError

        """

        _ = self.get_db_field_name

        assert labourer.id == task[_('labourer_id')], f"Task doesn't belong to the Labourer {labourer}: {task}"

        self.dynamo_db_client.update(
                {_('task_id'): task[_('task_id')]},
                attributes_to_update={_('greenfield'): int(time.time()) + self.config['greenfield_invocation_delta']},
                attributes_to_increment={_('attempts'): 1},
                condition_expression=f"{_('greenfield')} < {labourer.get_attr('start')}"
        )


    # Deprecated
    # def close_task(self, task_id: str, labourer_id: str):
    #     _ = self.get_db_field_name
    #
    #     self.dynamo_db_client.update(
    #             {_('task_id'): task_id, _('labourer_id'): labourer_id},
    #             attributes_to_update={_('closed_at'): int(time.time())},
    #     )

    def archive_task(self, task_id: str):
        _ = self.get_db_field_name

        # Get task
        task = self.get_task_by_id(task_id)

        # Update labourer_id_task_status field.
        is_completed = 1 if task.get(_('completed_at')) else 0
        labourer_id = task.get(_('labourer_id'))
        task[_('labourer_id_task_status')] = f"{labourer_id}_{is_completed}"
        task[_('closed_at')] = int(time.time())

        # Add it to completed tasks table:
        self.dynamo_db_client.put(task, table_name=self.config.get('sosw_closed_tasks_table'))

        # Delete it from tasks_table
        keys = {_('task_id'): task[_('task_id')]}
        self.dynamo_db_client.delete(keys)

        self.stats['archived_tasks'] += 1


    def get_task_by_id(self, task_id: str) -> Dict:
        """ Fetches the full data of the Task. """

        tasks = self.dynamo_db_client.get_by_query({self.get_db_field_name('task_id'): task_id})
        assert len(tasks) in [0, 1], "Fetched more than 1 task by primary key(). Something broke your DB " \
                                     "schema."
        return tasks[0] if tasks else {}


    def get_next_for_labourer(self, labourer: Labourer, cnt: int = 1, only_ids: bool = False) -> List[Union[str, Dict]]:
        """
        Fetch the next task(s) from the queue for the Labourer.

        :param labourer:    Labourer to get next tasks for.
        :param cnt:         Optional number of Tasks to fetch.
        :param only_ids:    If explicitly set True, then returns only the IDs of tasks.
                            This could save some transport if you are sending big batches of tasks between Lambdas.

        """

        # Maximum value to identify the task as available for invocation (either new, or ready for retry).
        max_greenfield = labourer.get_attr('start')

        result = self.dynamo_db_client.get_by_query(
                {
                    self.get_db_field_name('labourer_id'): labourer.id,
                    self.get_db_field_name('greenfield'):  max_greenfield
                },
                table_name=self.config['dynamo_db_config']['table_name'],
                index_name=self.config['dynamo_db_config']['index_greenfield'],
                fetch_all_fields=False,
                max_items=cnt,
                comparisons={
                    self.get_db_field_name('greenfield'): '<'
                })

        logger.debug(f"get_next_for_labourer() received: {result} from {self.config['dynamo_db_config']['table_name']} "
                     f"for labourer: {labourer.id} max greenfield: {max_greenfield}")

        return result if not only_ids else [task[self.get_db_field_name('task_id')] for task in result]


    def get_invoked_tasks_for_labourer(self, labourer: Labourer, completed: Optional[bool] = None) -> List[Dict]:
        """
        Return a list of tasks of current Labourer invoked during the current run of the Orchestrator.

        If completed is provided:
        * True - filter completed ones
        * False - filter NOT completed ones
        * None (default) - do not care about `completed` status.
        """

        _ = self.get_db_field_name

        query_args = {
            'keys':        {
                _('labourer_id'): labourer.id,
                _('greenfield'):  labourer.get_attr('invoked')
            },
            'comparisons': {_('greenfield'): '>='},
            'index_name':  self.config['dynamo_db_config']['index_greenfield'],
        }

        if completed is True:
            query_args['filter_expression'] = f"attribute_exists {_('completed_at')}"
        elif completed is False:
            query_args['filter_expression'] = f"attribute_not_exists {_('completed_at')}"
        else:
            logger.debug(f"No filtering by completed status for {query_args}")

        return self.dynamo_db_client.get_by_query(**query_args)


    def get_running_tasks_for_labourer(self, labourer: Labourer, count: bool = False) -> Union[List[Dict], int]:
        """
        Return a list of tasks of Labourer previously invoked, but not yet closed or expired.
        We assume they are still running.

        If `count` is specified as True will return just the number of tasks, not the items themselves.
        Much cheaper.
        """

        _ = self.get_db_field_name

        q = dict(
                keys={
                    _('labourer_id'):                labourer.id,
                    f"st_between_{_('greenfield')}": labourer.get_attr('expired'),
                    f"en_between_{_('greenfield')}": labourer.get_attr('invoked'),
                },
                index_name=self.config['dynamo_db_config']['index_greenfield'],
                filter_expression=f'attribute_not_exists {_("completed_at")}'
        )

        if count:
            q['return_count'] = True

        return self.dynamo_db_client.get_by_query(**q)


    def get_count_of_running_tasks_for_labourer(self, labourer: Labourer) -> int:
        """
        Returns a number of tasks we assume to be still running.
        Theoretically they can be dead with Exception, but not yet expired.
        """

        return self.get_running_tasks_for_labourer(labourer=labourer, count=True)


    def get_completed_tasks_for_labourer(self, labourer: Labourer) -> List[Dict]:
        """
        Return a list of tasks of the Labourer marked as completed.
        Scavenger is supposed to archive them all so no special filtering is required here.

        In order to be able to use the already existing `index_greenfield`, we sort tasks only in invoked stages
        (`greenfield > now()`). This number is supposed to be small, so filtering by an un-indexed field will be fast.
        """

        _ = self.get_db_field_name

        query_args = {
            'keys':              {
                _('labourer_id'): labourer.id,
                _('greenfield'):  str(time.time()),
            },
            'comparisons':       {_('greenfield'): '>='},
            'index_name':        self.config['dynamo_db_config']['index_greenfield'],
            'filter_expression': f"attribute_exists {_('completed_at')}",
        }

        return self.dynamo_db_client.get_by_query(**query_args)


    def get_expired_tasks_for_labourer(self, labourer: Labourer) -> List[Dict]:
        """ Return a list of tasks of Labourer previously invoked, and expired without being closed. """

        _ = self.get_db_field_name

        return self.dynamo_db_client.get_by_query(
                keys={
                    _('labourer_id'):                labourer.id,
                    f"st_between_{_('greenfield')}": labourer.get_attr('start'),
                    f"en_between_{_('greenfield')}": labourer.get_attr('expired'),
                },
                index_name=self.config['dynamo_db_config']['index_greenfield'],
                filter_expression=f"attribute_not_exists {_('completed_at')}",
        )


    def _jsonify_payload_of_task(self, task: Dict) -> Dict:
        """
        Simple helper to make sure the `payload` of the `task` is a string. If it's a dict - JSONify it.
        :param dict task:
        """

        payload = task.get('payload')
        if isinstance(payload, dict):
            logger.debug(f"JSON-ify payload of the task: {task}")
            task['payload'] = json.dumps(payload)

        return task


    def move_task_to_retry_table(self, task: Dict, wanted_delay: int):
        """
        Put the task to a Dynamo table `sosw_retry_tasks`, with the wanted delay: labourer.max_runtime * attempts.
        Delete it from `sosw_tasks` table.
        """

        _ = self.get_db_field_name

        # Add task to retry table
        retry_row = task.copy()
        retry_row[_('desired_launch_time')] = int(time.time()) + wanted_delay
        retry_row = self._jsonify_payload_of_task(retry_row)

        self.dynamo_db_client.put(retry_row, table_name=self.config.get('sosw_retry_tasks_table'))

        # Delete task from tasks table
        delete_keys = {_('task_id'): task[_('task_id')]}
        self.dynamo_db_client.delete(delete_keys)

        self.stats['scheduled_for_retry_later_tasks'] += 1


    def get_tasks_to_retry_for_labourer(self, labourer: Labourer, limit: int = None) -> List[Dict]:
        _ = self.get_db_field_name

        attrs = {
            'keys':        {_('labourer_id'): labourer.id, _('desired_launch_time'): str(labourer.get_attr('start'))},
            'comparisons': {_('desired_launch_time'): "<="},
            'table_name':  self.config['sosw_retry_tasks_table'],
            'index_name':  self.config['sosw_retry_tasks_greenfield_index'],
        }
        if limit:
            attrs['max_items'] = limit
        tasks = self.dynamo_db_client.get_by_query(**attrs)
        return tasks


    def retry_tasks(self, labourer: Labourer, tasks: List[Dict]):
        """
        Move tasks to tasks table, in beginning of the queue (with greenfield of a task that will be invoked next)
        All tasks must belong to the same labourer.
        """

        _ = self.get_db_field_name

        for task in tasks:
            assert task[_('labourer_id')] == labourer.id, f"Task labourer_id must be {labourer.id}, " \
                f"bad value: {task[_('labourer_id')]}"

        lowest_greenfield = self.get_oldest_greenfield_for_labourer(labourer)

        for task in tasks:
            del task['desired_launch_time']
            lowest_greenfield = lowest_greenfield - 1
            task[_('greenfield')] = lowest_greenfield
            task = self._jsonify_payload_of_task(task)

            delete_keys = {_('labourer_id'): labourer.id, _('task_id'): task[_('task_id')]}

            # If boto supports DynamoDB transaction, use them to add task to tasks_table and delete from retry_table
            # https://github.com/boto/boto3/issues/1791: It's available for 1.9.54+
            if parse_version(str(boto3.__version__)) >= parse_version('1.9.54'):
                put_query = self.dynamo_db_client.make_put_transaction_item(task)
                delete_query = self.dynamo_db_client.make_delete_transaction_item(
                        delete_keys, table_name=self.config.get('sosw_retry_tasks_table'))
                self.dynamo_db_client.transact_write(put_query, delete_query)

            else:
                logger.info("Looks like you are running an ancient copy of boto3 still in old Environment of Lambda."
                            "Salut to AWS from March 2019.")
                self.dynamo_db_client.put(task)
                self.dynamo_db_client.delete(keys=delete_keys, table_name=self.config.get('sosw_retry_tasks_table'))

            self.stats['due_for_retry_tasks'] += 1


    @benchmark
    def get_average_labourer_duration(self, labourer: Labourer) -> int:
        """
        Analyse latest tasks of Labourer and calculate average runtime duration.

        .. warning:: This method doesn't know the exact duration of failed attempts.
                     Thus if the task is completely failed, we assume that all attempts failed at maximum duration.

        :return:    Average duration in seconds.
        """

        _ = self.get_db_field_name
        _cfg = self.config.get

        durations = []

        q = dict(
                keys={
                    _('labourer_id_task_status'): f"{labourer.id}_1",
                },
                table_name=_cfg('sosw_closed_tasks_table'),
                index_name=_cfg('sosw_closed_tasks_labourer_status_index'),
                max_items=_cfg('max_closed_to_analyse_for_duration'),
                desc=True
        )

        # Fetch last X closed tasks
        tasks = self.dynamo_db_client.get_by_query(**q)

        # Fetch failed tasks as well
        q['keys'][_('labourer_id_task_status')] = f"{labourer.id}_0"
        tasks.extend(self.dynamo_db_client.get_by_query(**q))

        # Now take the really last 50 ordered by greenfield (last invocation)
        tasks = sorted(tasks, key=lambda x: x.get(_('greenfield')))[:_cfg('max_closed_to_analyse_for_duration')]

        # Get their duration
        for task in tasks:
            # We assume duration of failed tasks to be maximum.
            if not task.get(_('completed_at')):
                durations.extend([labourer.get_attr('max_duration') for _ in range(int(task[_('attempts')]))])
            else:
                # Duration of completed tasks we calculate based on the value of last `greenfield` and `completed_at`
                durations.append(
                        task[_('completed_at')] - task[_('greenfield')] + _cfg('greenfield_invocation_delta'))

        # Return the average
        try:
            return round(sum(durations) / len(durations))
        except ZeroDivisionError:
            return 0



import boto3
import os
import unittest

from copy import deepcopy
from unittest import mock
from unittest.mock import MagicMock, patch

from sosw.orchestrator import Orchestrator
from sosw.labourer import Labourer
from sosw.test.variables import TEST_ORCHESTRATOR_CONFIG


os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class Orchestrator_UnitTestCase(unittest.TestCase):
    TEST_CONFIG = TEST_ORCHESTRATOR_CONFIG

    LABOURER = Labourer(id='some_function', arn='arn:aws:lambda:us-west-2:000000000000:function:some_function')

    def setUp(self):
        self.patcher = patch("sosw.app.get_config")
        self.get_config_patch = self.patcher.start()

        self.custom_config = deepcopy(self.TEST_CONFIG)
        with patch('boto3.client'):
            self.orchestrator = Orchestrator(self.custom_config)



    def tearDown(self):
        self.patcher.stop()

        try:
            del (os.environ['AWS_LAMBDA_FUNCTION_NAME'])
        except Exception:
            pass


    def test_true(self):
        self.assertEqual(1, 1)


    # @unittest.skip("Depricated method")
    # def test_get_labourer_setting(self):
    #
    #     custom_config = self.TEST_CONFIG.copy()
    #     custom_config['labourers'] = {
    #         42: {'foo': 'bar'},
    #     }
    #
    #     with patch('boto3.client'):
    #         orchestrator = Orchestrator(custom_config)
    #
    #     self.assertEqual(orchestrator.get_labourer_setting(Labourer(id=42), 'foo'), 'bar')
    #     self.assertEqual(orchestrator.get_labourer_setting(Labourer(id=42), 'faz'), None)
    #
    #     self.assertEqual(orchestrator.get_labourer_setting(Labourer(id=4422), 'foo'), None)
    #     self.assertEqual(orchestrator.get_labourer_setting(Labourer(id=4422), 'faz'), None)


    def test_get_desired_invocation_number_for_labourer(self):

        # Status - expected output for max invocations = 10
        TESTS = {
            0: 0,
            1: 0,
            2: 5,
            3: 7,
            4: 10
        }

        some_labourer = self.orchestrator.task_client.register_labourers()[0]

        with patch('boto3.client'):
            orchestrator = Orchestrator(self.custom_config)

        orchestrator.task_client = MagicMock()

        print(some_labourer.get_attr('max_simultaneous_invocations'))
        for eco, expected in TESTS.items():
            orchestrator.task_client.ecology_client.get_labourer_status.return_value = eco
            orchestrator.task_client.ecology_client.count_running_tasks_for_labourer.return_value = 0
            self.assertEqual(orchestrator.get_desired_invocation_number_for_labourer(some_labourer), expected)


    def test_get_desired_invocation_number_for_labourer__default(self):

        # Status - expected output for max invocations = 2
        TESTS = {
            0: 0,
            1: 0,
            2: 5,
            3: 7,
            4: 10
        }

        # self.orchestrator.task_client.register_labourers.return_value = [self.LABOURER]
        # self.orchestrator.task_client.ecology_client.count_running_tasks_for_labourer.return_value = 0

        some_labourer = self.orchestrator.task_client.register_labourers()[0]

        # Once registered Labourers we Mock the task client.
        self.orchestrator.task_client = MagicMock()
        self.orchestrator.task_client.ecology_client.count_running_tasks_for_labourer.return_value = 0

        for eco, expected in TESTS.items():
            self.orchestrator.task_client.ecology_client.get_labourer_status.return_value = eco

            self.assertEqual(self.orchestrator.get_desired_invocation_number_for_labourer(some_labourer), expected)


    def test_invoke_for_labourer(self):
        TEST_COUNT = 3
        some_labourer = self.orchestrator.task_client.register_labourers()[0]

        self.orchestrator.get_desired_invocation_number_for_labourer = MagicMock(return_value=TEST_COUNT)

        self.orchestrator.invoke_for_labourer(some_labourer)

        self.orchestrator.get_desired_invocation_number_for_labourer.assert_called_once()


    def test_invoke_for_labourer__desired_zero(self):
        self.orchestrator.get_desired_invocation_number_for_labourer = MagicMock(return_value=0)
        self.orchestrator.task_client.invoke_task = MagicMock()

        self.orchestrator.invoke_for_labourer(self.LABOURER)

        self.orchestrator.task_client.invoke_task.assert_not_called()



__all__ = ['extract_call_params', 'line_count']

import inspect
import subprocess

from typing import Dict


def extract_call_params(call_args, function) -> Dict:
    """
    Combines a dictionary out of Mock.call_args.
    Helper useful to validate specific call parameters of the Mocked function.

    When you are not sure if your function is called with args or kwargs, you just feed the call_args
    and the source of function to this helper and receive a dictionary.

    .. _code-block:: python

       call_kwargs = extract_call_params(your_mock_object.some_method.call_args, mocked_module.Class.method)

    :param call_args:   call_args of Mock object
    :param function:    Source object that was initially mocked
    :rtype:             dict
    """

    # Specification of arguments of function
    function_args = inspect.getfullargspec(function).args

    # Mock call_arguments as a tuple
    call_args, call_kwargs = call_args

    result = {}

    for i, v in enumerate(call_args):
        position = i + 1 if function_args[0] == 'self' else i
        result[function_args[position]] = v

    for k, v in call_kwargs.items():
        result[k] = v

    return result


def line_count(file):
    try:
        # This is the FAST way
        return int(subprocess.check_output(f'wc -l {file}', shell=True).split()[0])
    except Exception:
        # This is in case you are running in some creepy environment without shell access
        i = 0
        with open(file, 'r') as f:
            for i, _ in enumerate(f):
                pass
        return i + 1



# import logging
# import unittest
# import os
# import boto3
#
# from unittest.mock import patch, MagicMock
#
# from sosw.components.tasks_api_client_for_workers import *
#
#
# logger = logging.getLogger()
#
# os.environ["STAGE"] = "test"
# os.environ["autotest"] = "True"
#
# TASK_ID = 'test_task_id'
# LABOURER_ID = 'some_lambda'
# TABLE_NAME = 'autotest_sosw_tasks'
#
#
# class tasks_api_client_for_workers_TestCase(unittest.TestCase):
#
#     def setUp(self):
#         """ Setting up an item in dynamo DB """
#
#         self.patcher = patch("boto3.client")
#         self.boto3_client_patch = self.patcher.start()
#
#
#     def tearDown(self):
#         self.patcher.stop()
#
#
#     @patch('os.environ.get')
#     def test_close_task__laborer_id_from_end(self, env_mock):
#         """ Careful! Once we patch os.environ - we are no longer in STAGE=test! :) """
#
#         env_mock.return_value = LABOURER_ID
#         mock_client = MagicMock()
#         self.boto3_client_patch.return_value = mock_client
#
#         close_task({'task_id': TASK_ID})
#
#         mock_client.update_item.assert_called_once()
#
#         args, kwargs = mock_client.update_item.call_args
#         # You may do some validation here.
#
#
# if __name__ == '__main__':
#     unittest.main()



from setuptools import setup, find_packages


with open("README.md", "r") as f:
    long_description = f.read()

setup(name='sosw',
      version='0.7.15',
      description='Serverless Orchestrator of Serverless Workers',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://github.com/bimpression/sosw',
      author='Nikolay Grishchenko',
      author_email='nikolay@bimpression.com',
      license='MIT',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Operating System :: Other OS',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Software Development'
      ],
      # packages=['sosw'],
      packages=find_packages(exclude=['docs', 'test']),
      install_requires=[
          'boto3>=1.9'
      ],
      zip_safe=False)



import boto3
import os
import random
import types
import unittest

from unittest.mock import MagicMock, patch

from sosw.scheduler import Scheduler, global_vars
from sosw.labourer import Labourer
from sosw.test.variables import TEST_SCHEDULER_CONFIG
from sosw.test.helpers_test import line_count

os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class Scheduler_IntegrationTestCase(unittest.TestCase):
    TEST_CONFIG = TEST_SCHEDULER_CONFIG

    @classmethod
    def setUpClass(cls):
        cls.TEST_CONFIG['init_clients'] = ['S3', ]

        cls.clean_bucket()


    @staticmethod
    def clean_bucket():
        """ Clean S3 bucket"""

        s3 = boto3.resource('s3')
        bucket = s3.Bucket('autotest-bucket')
        bucket.objects.all().delete()


    def exists_in_s3(self, key):
        try:
            self.s3_client.get_object(Bucket='autotest-bucket', Key=key)
            return True
        except self.s3_client.exceptions.ClientError:
            return False


    def put_file(self, local=None, key=None, only_remote=False):
        self.make_local_file('Liat')

        self.s3_client.upload_file(Filename=local or self.scheduler.local_queue_file,
                                   Bucket='autotest-bucket',
                                   Key=key or self.scheduler.remote_queue_file)

        if only_remote:
            try:
                os.remove(local or self.scheduler.local_queue_file)
            except Exception:
                pass


    def make_local_file(self, girl_name="Athena", fname=None, rows=10):

        with open(fname or self.scheduler.local_queue_file, 'w') as f:
            for i in range(rows):
                f.write(f"hello {girl_name} {i} {random.randint(0,99)}\n")


    def setUp(self):
        self.patcher = patch("sosw.app.get_config")
        self.get_config_patch = self.patcher.start()

        self.custom_config = self.TEST_CONFIG.copy()
        lambda_context = types.SimpleNamespace()
        lambda_context.aws_request_id = 'AWS_REQ_ID'
        lambda_context.invoked_function_arn = 'arn:aws:lambda:us-west-2:000000000000:function:some_function'
        lambda_context.get_remaining_time_in_millis = MagicMock(side_effect=[100000, 100])
        global_vars.lambda_context = lambda_context

        self.scheduler = Scheduler(self.custom_config)

        self.s3_client = boto3.client('s3')


    def tearDown(self):
        self.patcher.stop()
        self.clean_bucket()

        try:
            os.remove(self.scheduler.local_queue_file)
        except Exception:
            pass

        try:
            del (os.environ['AWS_LAMBDA_FUNCTION_NAME'])
        except Exception:
            pass


    def test_get_and_lock_queue_file(self):
        self.put_file(only_remote=True)

        # Check old artifacts
        self.assertFalse(self.exists_in_s3(self.scheduler.remote_queue_locked_file))
        self.assertTrue(self.exists_in_s3(self.scheduler.remote_queue_file))

        r = self.scheduler.get_and_lock_queue_file()

        self.assertEqual(r, self.scheduler.local_queue_file)

        self.assertTrue(self.exists_in_s3(self.scheduler.remote_queue_locked_file))
        self.assertFalse(self.exists_in_s3(self.scheduler.remote_queue_file))

        number_of_lines = line_count(self.scheduler.local_queue_file)
        # print(f"Number of lines: {number_of_lines}")
        self.assertTrue(number_of_lines, 10)


    def test_upload_and_unlock_queue_file(self):

        # Check old artifacts
        self.assertFalse(self.exists_in_s3(self.scheduler.remote_queue_locked_file))
        self.assertFalse(self.exists_in_s3(self.scheduler.remote_queue_file))

        self.make_local_file('Demida')

        self.scheduler.upload_and_unlock_queue_file()

        self.assertFalse(self.exists_in_s3(self.scheduler.remote_queue_locked_file))
        self.assertTrue(self.exists_in_s3(self.scheduler.remote_queue_file))


    def test_upload_and_unlock_queue_file__handles_existing_locked(self):

        self.put_file(key=self.scheduler.remote_queue_locked_file)

        # Check old artifacts
        self.assertTrue(self.exists_in_s3(self.scheduler.remote_queue_locked_file))
        self.assertFalse(self.exists_in_s3(self.scheduler.remote_queue_file))

        self.make_local_file('Nora')

        self.scheduler.upload_and_unlock_queue_file()

        self.assertFalse(self.exists_in_s3(self.scheduler.remote_queue_locked_file))
        self.assertTrue(self.exists_in_s3(self.scheduler.remote_queue_file))



import boto3
import os
import unittest

from copy import deepcopy
from unittest.mock import patch

from ..app import Processor
from ..components.sns import SnsManager
from ..components.siblings import SiblingsManager


os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class app_TestCase(unittest.TestCase):
    TEST_CONFIG = {
        'test':            True,
        'siblings_config': {'test': True},
    }


    def setUp(self):
        self.patcher = patch("sosw.app.get_config")
        self.get_config_patch = self.patcher.start()

        config = deepcopy(self.TEST_CONFIG)
        self.processor = Processor(custom_config=config)


    def tearDown(self):
        self.patcher.stop()


    def test_app_init(self):
        self.assertTrue(True)


    def test_app_init__with_some_clients(self):
        custom_config = deepcopy(self.TEST_CONFIG)
        custom_config.update({
            'init_clients': ['Sns', 'Siblings']
        })

        processor = Processor(custom_config=custom_config)
        self.assertIsInstance(getattr(processor, 'sns_client'), SnsManager,
                              "SnsManager was not initialized. Probably boto3 sns instead of it.")
        self.assertIsNotNone(getattr(processor, 'siblings_client'))


    def test_app_init__boto_and_components_custom_clients(self):
        custom_config = deepcopy(self.TEST_CONFIG)
        custom_config.update({
            'init_clients': ['lambda', 'Siblings']
        })

        processor = Processor(custom_config=custom_config)
        self.assertIsInstance(getattr(processor, 'siblings_client'), SiblingsManager)

        # Clients of boto3 will not be exactly of same type (something dynamic in boto3), so we can't compare classes.
        # Let us assume that checking the class_name is enough for this test.
        self.assertEqual(str(type(getattr(processor, 'lambda_client'))), str(type(boto3.client('lambda'))))


    def test_app_init__with_some_invalid_client(self):
        custom_config = {
            'init_clients': ['NotExists', 'Sns']
        }
        self.assertRaises(RuntimeError, Processor, custom_config=custom_config)



import boto3
import logging
import time
import unittest
import os


logging.getLogger('botocore').setLevel(logging.WARNING)

os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"

from sosw.components.dynamo_db import DynamoDbClient, clean_dynamo_table
from sosw.components.helpers import chunks


class dynamodb_client_IntegrationTestCase(unittest.TestCase):
    TEST_CONFIG = {
        'row_mapper':      {
            'lambda_name':   'S',
            'invocation_id': 'S',
            'en_time':       'N',

            'hash_col':      'S',
            'range_col':     'N',
            'other_col':     'S',
            'new_col':       'S',
            'some_col':      'S',
            'some_counter':  'N',
            'some_bool':     'BOOL',
        },
        'required_fields': ['lambda_name'],
        'table_name':      'autotest_dynamo_db',
        'hash_key': 'hash_col'
    }


    @classmethod
    def setUpClass(cls):
        clean_dynamo_table()


    def setUp(self):
        self.HASH_COL = 'hash_col'
        self.HASH_KEY = (self.HASH_COL, 'S')

        self.RANGE_COL = 'range_col'
        self.RANGE_COL_TYPE = 'N'
        self.RANGE_KEY = (self.RANGE_COL, self.RANGE_COL)

        self.KEYS = (self.HASH_COL, self.RANGE_COL)
        self.table_name = 'autotest_dynamo_db'
        self.dynamo_client = DynamoDbClient(config=self.TEST_CONFIG)


    def tearDown(self):
        clean_dynamo_table(self.table_name, self.KEYS)


    def test_put(self):
        row = {self.HASH_COL: 'cat', self.RANGE_COL: '123', 'some_bool': True}

        client = boto3.client('dynamodb')

        client.delete_item(TableName=self.table_name,
                           Key={
                               self.HASH_COL:  {'S': str(row[self.HASH_COL])},
                               self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])},
                           })

        self.dynamo_client.put(row, self.table_name)

        result = client.scan(TableName=self.table_name,
                             FilterExpression="hash_col = :hash_col AND range_col = :range_col",
                             ExpressionAttributeValues={
                                 ':hash_col':  {'S': row[self.HASH_COL]},
                                 ':range_col': {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                             }
                             )

        items = result['Items']

        self.assertEqual(1, len(items))

        expected = [{'hash_col':  {'S': 'cat'},
                     'range_col': {'N': '123'},
                     'some_bool': {'BOOL': True}}]
        self.assertEqual(expected, items)


    def test_put__create(self):
        row = {self.HASH_COL: 'cat', self.RANGE_COL: '123'}

        self.dynamo_client.put(row, self.table_name)

        with self.assertRaises(self.dynamo_client.dynamo_client.exceptions.ConditionalCheckFailedException):
            self.dynamo_client.put(row, self.table_name, overwrite_existing=False)


    def test_update__updates(self):
        keys = {self.HASH_COL: 'cat', self.RANGE_COL: '123'}
        row = {self.HASH_COL: 'cat', self.RANGE_COL: '123', 'some_col': 'no', 'other_col': 'foo'}
        attributes_to_update = {'some_col': 'yes', 'new_col': 'yup'}

        self.dynamo_client.put(row, self.table_name)

        client = boto3.client('dynamodb')

        # First check that the row we are trying to update is PUT correctly.
        initial_row = client.get_item(
                Key={
                    self.HASH_COL:  {'S': row[self.HASH_COL]},
                    self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                },
                TableName=self.table_name,
        )['Item']

        initial_row = self.dynamo_client.dynamo_to_dict(initial_row)

        self.assertIsNotNone(initial_row)
        self.assertEqual(initial_row['some_col'], 'no')
        self.assertEqual(initial_row['other_col'], 'foo')

        self.dynamo_client.update(keys, attributes_to_update, table_name=self.table_name)

        updated_row = client.get_item(
                Key={
                    self.HASH_COL:  {'S': row[self.HASH_COL]},
                    self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                },
                TableName=self.table_name,
        )['Item']

        updated_row = self.dynamo_client.dynamo_to_dict(updated_row)

        self.assertIsNotNone(updated_row)
        self.assertEqual(updated_row['some_col'], 'yes'), "Updated field not really updated"
        self.assertEqual(updated_row['new_col'], 'yup'), "New field was not created"
        self.assertEqual(updated_row['other_col'], 'foo'), "This field should be preserved, update() damaged it"


    def test_update__increment(self):
        keys = {self.HASH_COL: 'cat', self.RANGE_COL: '123'}
        row = {self.HASH_COL: 'cat', self.RANGE_COL: '123', 'some_col': 'no', 'some_counter': 10}
        attributes_to_increment = {'some_counter': '1'}

        self.dynamo_client.put(row, self.table_name)

        self.dynamo_client.update(keys, {}, attributes_to_increment=attributes_to_increment, table_name=self.table_name)

        client = boto3.client('dynamodb')

        updated_row = client.get_item(
                Key={
                    self.HASH_COL:  {'S': row[self.HASH_COL]},
                    self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                },
                TableName=self.table_name,
        )['Item']

        updated_row = self.dynamo_client.dynamo_to_dict(updated_row)

        self.assertIsNotNone(updated_row)
        self.assertEqual(updated_row['some_counter'], 11)


    def test_update__increment_2(self):
        keys = {self.HASH_COL: 'cat', self.RANGE_COL: '123'}
        row = {self.HASH_COL: 'cat', self.RANGE_COL: '123', 'some_col': 'no', 'some_counter': 10}
        attributes_to_increment = {'some_counter': 5}

        self.dynamo_client.put(row, self.table_name)

        self.dynamo_client.update(keys, {}, attributes_to_increment=attributes_to_increment, table_name=self.table_name)

        client = boto3.client('dynamodb')

        updated_row = client.get_item(
                Key={
                    self.HASH_COL:  {'S': row[self.HASH_COL]},
                    self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                },
                TableName=self.table_name,
        )['Item']

        updated_row = self.dynamo_client.dynamo_to_dict(updated_row)

        self.assertIsNotNone(updated_row)
        self.assertEqual(updated_row['some_counter'], 15)


    def test_update__increment_no_default(self):
        keys = {self.HASH_COL: 'cat', self.RANGE_COL: '123'}
        row = {self.HASH_COL: 'cat', self.RANGE_COL: '123', 'some_col': 'no'}
        attributes_to_increment = {'some_counter': '3'}

        self.dynamo_client.put(row, self.table_name)

        self.dynamo_client.update(keys, {}, attributes_to_increment=attributes_to_increment, table_name=self.table_name)

        client = boto3.client('dynamodb')

        updated_row = client.get_item(
                Key={
                    self.HASH_COL:  {'S': row[self.HASH_COL]},
                    self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                },
                TableName=self.table_name,
        )['Item']

        updated_row = self.dynamo_client.dynamo_to_dict(updated_row)

        self.assertIsNotNone(updated_row)
        self.assertEqual(updated_row['some_counter'], 3)


    def test_update__condition_expression(self):
        keys = {self.HASH_COL: 'slime', self.RANGE_COL: '41'}
        row = {self.HASH_COL: 'slime', self.RANGE_COL: '41', 'some_col': 'no'}

        self.dynamo_client.put(row, self.table_name)

        # Should fail because conditional expression does not match
        self.assertRaises(self.dynamo_client.dynamo_client.exceptions.ConditionalCheckFailedException,
                          self.dynamo_client.update, keys, {},
                          attributes_to_increment={'some_counter': '3'},
                          condition_expression='some_col = yes', table_name=self.table_name)

        # Should pass
        self.dynamo_client.update(keys, {}, attributes_to_increment={'some_counter': '3'},
                                  condition_expression='some_col = no', table_name=self.table_name)

        client = boto3.client('dynamodb')
        updated_row = client.get_item(
                Key={
                    self.HASH_COL:  {'S': row[self.HASH_COL]},
                    self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                },
                TableName=self.table_name,
        )['Item']

        updated_row = self.dynamo_client.dynamo_to_dict(updated_row)
        self.assertEqual(updated_row['some_counter'], 3)


    def test_patch(self):
        keys = {self.HASH_COL: 'slime', self.RANGE_COL: '41'}
        row = {self.HASH_COL: 'slime', self.RANGE_COL: '41', 'some_col': 'no'}

        # Should fail because row doesn't exist
        self.assertRaises(self.dynamo_client.dynamo_client.exceptions.ConditionalCheckFailedException,
                          self.dynamo_client.patch, keys, attributes_to_update={'some_col': 'yes'}, table_name=self.table_name)

        # Create the row
        self.dynamo_client.put(row, self.table_name)
        # Should pass because the row exists now
        self.dynamo_client.patch(keys, attributes_to_update={'some_col': 'yes'}, table_name=self.table_name)

        client = boto3.client('dynamodb')
        updated_row = client.get_item(
                Key={
                    self.HASH_COL:  {'S': row[self.HASH_COL]},
                    self.RANGE_COL: {self.RANGE_COL_TYPE: str(row[self.RANGE_COL])}
                },
                TableName=self.table_name,
        )['Item']

        updated_row = self.dynamo_client.dynamo_to_dict(updated_row)
        self.assertEqual(updated_row['some_col'], 'yes')


    def test_get_by_query__primary_index(self):
        keys = {self.HASH_COL: 'cat', self.RANGE_COL: '123'}
        row = {self.HASH_COL: 'cat', self.RANGE_COL: 123, 'some_col': 'test', 'some_bool': True}
        self.dynamo_client.put(row, self.table_name)

        result = self.dynamo_client.get_by_query(keys=keys)

        self.assertEqual(len(result), 1)
        result = result[0]
        for key in row:
            self.assertEqual(row[key], result[key])
        for key in result:
            self.assertEqual(row[key], result[key])


    def test_get_by_query__primary_index__gets_multiple(self):
        row = {self.HASH_COL: 'cat', self.RANGE_COL: 123, 'some_col': 'test'}
        self.dynamo_client.put(row, self.table_name)

        row2 = {self.HASH_COL: 'cat', self.RANGE_COL: 1234, 'some_col': 'test2'}
        self.dynamo_client.put(row2, self.table_name)

        result = self.dynamo_client.get_by_query(keys={self.HASH_COL: 'cat'})

        self.assertEqual(len(result), 2)

        result1 = [x for x in result if x[self.RANGE_COL] == row[self.RANGE_COL]][0]
        result2 = [x for x in result if x[self.RANGE_COL] == row2[self.RANGE_COL]][0]

        for key in row:
            self.assertEqual(row[key], result1[key])
        for key in result1:
            self.assertEqual(row[key], result1[key])
        for key in row2:
            self.assertEqual(row2[key], result2[key])
        for key in result2:
            self.assertEqual(row2[key], result2[key])


    def test_get_by_query__secondary_index(self):
        keys = {self.HASH_COL: 'cat', 'other_col': 'abc123'}
        row = {self.HASH_COL: 'cat', self.RANGE_COL: 123, 'other_col': 'abc123'}
        self.dynamo_client.put(row, self.table_name)

        result = self.dynamo_client.get_by_query(keys=keys, index_name='autotest_index')

        self.assertEqual(len(result), 1)
        result = result[0]
        for key in row:
            self.assertEqual(row[key], result[key])
        for key in result:
            self.assertEqual(row[key], result[key])


    def test_get_by_query__comparison(self):
        keys = {self.HASH_COL: 'cat', self.RANGE_COL: '300'}
        row1 = {self.HASH_COL: 'cat', self.RANGE_COL: 123, 'other_col': 'abc123'}
        row2 = {self.HASH_COL: 'cat', self.RANGE_COL: 456, 'other_col': 'abc123'}
        self.dynamo_client.put(row1, self.table_name)
        self.dynamo_client.put(row2, self.table_name)

        result = self.dynamo_client.get_by_query(keys=keys, comparisons={self.RANGE_COL: '<='})

        self.assertEqual(len(result), 1)

        result = result[0]
        self.assertEqual(result, row1)


    def test_get_by_query__comparison_between(self):
        # Put sample data
        x = [self.dynamo_client.put({self.HASH_COL: 'cat', self.RANGE_COL: x}, self.table_name) for x in range(10)]

        keys = {self.HASH_COL: 'cat', 'st_between_range_col': '3', 'en_between_range_col': '6'}
        result = self.dynamo_client.get_by_query(keys=keys, comparisons={self.RANGE_COL: 'between'})
        # print(result)
        self.assertTrue(all(x[self.RANGE_COL] in range(3, 7) for x in result))

        result = self.dynamo_client.get_by_query(keys=keys)
        # print(result)
        self.assertTrue(all(x[self.RANGE_COL] in range(3, 7) for x in result)), "Failed if unspecified comparison. " \
                                                                                "Should be automatic for :st_between_..."


    def test_get_by_query__filter_expression(self):
        """
        This _integration_ test runs multiple checks with same sample data for several comparators.
        Have a look at the manual if required:
        https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.OperatorsAndFunctions.html
        """

        # Put sample data
        [self.dynamo_client.put({self.HASH_COL: 'cat', self.RANGE_COL: x}, self.table_name) for x in range(3)]
        [self.dynamo_client.put({self.HASH_COL: 'cat', self.RANGE_COL: x, 'mark': 1}, self.table_name) for x in
         range(3, 6)]
        self.dynamo_client.put({self.HASH_COL: 'cat', self.RANGE_COL: 6, 'mark': 0}, self.table_name)
        self.dynamo_client.put({self.HASH_COL: 'cat', self.RANGE_COL: 7, 'mark': 'a'}, self.table_name)

        # Condition by range_col will return five rows out of six: 0 - 4
        # Filter expression neggs the first three rows because they don't have `mark = 1`.
        keys = {self.HASH_COL: 'cat', self.RANGE_COL: 4}
        result = self.dynamo_client.get_by_query(keys=keys, comparisons={self.RANGE_COL: '<='},
                                                 fetch_all_fields=True, filter_expression='mark = 1')
        # print(result)

        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], {self.HASH_COL: 'cat', self.RANGE_COL: 3, 'mark': 1})
        self.assertEqual(result[1], {self.HASH_COL: 'cat', self.RANGE_COL: 4, 'mark': 1})

        # In the same test we check also some comparator _functions_.
        result = self.dynamo_client.get_by_query(keys=keys, comparisons={self.RANGE_COL: '<='},
                                                 fetch_all_fields=True, filter_expression='attribute_exists mark')
        # print(result)
        self.assertEqual(len(result), 2)
        self.assertEqual([x[self.RANGE_COL] for x in result], list(range(3, 5)))

        self.assertEqual(result[0], {self.HASH_COL: 'cat', self.RANGE_COL: 3, 'mark': 1})
        self.assertEqual(result[1], {self.HASH_COL: 'cat', self.RANGE_COL: 4, 'mark': 1})

        result = self.dynamo_client.get_by_query(keys=keys, comparisons={self.RANGE_COL: '<='},
                                                 fetch_all_fields=True, filter_expression='attribute_not_exists mark')
        # print(result)
        self.assertEqual(len(result), 3)
        self.assertEqual([x[self.RANGE_COL] for x in result], list(range(3)))


    def test_get_by_query__comparison_begins_with(self):
        self.table_name = 'autotest_config_component'  # This table has a string range key
        self.HASH_KEY = ('env', 'S')
        self.RANGE_KEY = ('config_name', 'S')
        self.KEYS = ('env', 'config_name')
        config = {
            'row_mapper':      {
                'env':          'S',
                'config_name':  'S',
                'config_value': 'S'
            },
            'required_fields': ['env', 'config_name', 'config_value'],
            'table_name':      'autotest_config_component',
            'hash_key': self.HASH_COL
        }

        self.dynamo_client = DynamoDbClient(config=config)

        row1 = {'env': 'cat', 'config_name': 'testzing', 'config_value': 'abc123'}
        row2 = {'env': 'cat', 'config_name': 'dont_get_this', 'config_value': 'abc123'}
        row3 = {'env': 'cat', 'config_name': 'testzer', 'config_value': 'abc124'}
        self.dynamo_client.put(row1, self.table_name)
        self.dynamo_client.put(row2, self.table_name)
        self.dynamo_client.put(row3, self.table_name)

        keys = {'env': 'cat', 'config_name': 'testz'}
        result = self.dynamo_client.get_by_query(
                keys=keys,
                table_name=self.table_name,
                comparisons={'config_name': 'begins_with'}
        )

        self.assertEqual(len(result), 2)

        self.assertTrue(row1 in result)
        self.assertTrue(row3 in result)


    def test_get_by_query__max_items(self):
        # This function can also be used for some benchmarking, just change to bigger amounts manually.
        INITIAL_TASKS = 5  # Change to 500 to run benchmarking, and uncomment raise at the end of the test.

        for x in range(1000, 1000 + INITIAL_TASKS):
            row = {self.HASH_COL: f"key", self.RANGE_COL: x}
            self.dynamo_client.put(row, self.table_name)
            if INITIAL_TASKS > 10:
                time.sleep(0.1)  # Sleep a little to fit the Write Capacity (10 WCU) of autotest table.

        n = 3
        st = time.perf_counter()
        result = self.dynamo_client.get_by_query({self.HASH_COL: 'key'}, table_name=self.table_name, max_items=n)
        bm = time.perf_counter() - st
        logging.info(f"Benchmark (n={n}): {bm}")

        self.assertEqual(len(result), n)
        self.assertLess(bm, 0.1)

        # Check unspecified limit.
        result = self.dynamo_client.get_by_query({self.HASH_COL: 'key'}, table_name=self.table_name)
        self.assertEqual(len(result), INITIAL_TASKS)


    def test_get_by_query__return_count(self):
        rows = [
            {self.HASH_COL: 'cat1', self.RANGE_COL: 121, 'some_col': 'test1'},
            {self.HASH_COL: 'cat1', self.RANGE_COL: 122, 'some_col': 'test2'},
            {self.HASH_COL: 'cat1', self.RANGE_COL: 123, 'some_col': 'test3'}
        ]

        for x in rows:
            self.dynamo_client.put(x, table_name=self.table_name)

        result = self.dynamo_client.get_by_query({self.HASH_COL: 'cat1'}, table_name=self.table_name, return_count=True)

        self.assertEqual(result, 3)


    def test_get_by_query__reverse(self):
        rows = [
            {self.HASH_COL: 'cat1', self.RANGE_COL: 121, 'some_col': 'test1'},
            {self.HASH_COL: 'cat1', self.RANGE_COL: 122, 'some_col': 'test2'},
            {self.HASH_COL: 'cat1', self.RANGE_COL: 123, 'some_col': 'test3'}
        ]

        for x in rows:
            self.dynamo_client.put(x, table_name=self.table_name)

        result = self.dynamo_client.get_by_query({self.HASH_COL: 'cat1'}, table_name=self.table_name, desc=True)

        self.assertEqual(result[0], rows[-1])


    def test_get_by_scan__all(self):
        rows = [
            {self.HASH_COL: 'cat1', self.RANGE_COL: 121, 'some_col': 'test1'},
            {self.HASH_COL: 'cat2', self.RANGE_COL: 122, 'some_col': 'test2'},
            {self.HASH_COL: 'cat3', self.RANGE_COL: 123, 'some_col': 'test3'}
        ]
        for x in rows:
            self.dynamo_client.put(x, self.table_name)

        result = self.dynamo_client.get_by_scan()

        self.assertEqual(len(result), 3)

        for r in rows:
            assert r in result, f"row not in result from dynamo scan: {r}"


    def test_get_by_scan__with_filter(self):
        rows = [
            {self.HASH_COL: 'cat1', self.RANGE_COL: 121, 'some_col': 'test1'},
            {self.HASH_COL: 'cat1', self.RANGE_COL: 122, 'some_col': 'test2'},
            {self.HASH_COL: 'cat2', self.RANGE_COL: 122, 'some_col': 'test2'},
        ]
        for x in rows:
            self.dynamo_client.put(x, self.table_name)

        filter = {'some_col': 'test2'}

        result = self.dynamo_client.get_by_scan(attrs=filter)

        self.assertEqual(len(result), 2)

        for r in rows[1:]:
            assert r in result, f"row not in result from dynamo scan: {r}"


    def test_batch_get_items(self):
        rows = [
            {self.HASH_COL: 'cat1', self.RANGE_COL: 121, 'some_col': 'test1'},
            {self.HASH_COL: 'cat1', self.RANGE_COL: 122, 'some_col': 'test2'},
            {self.HASH_COL: 'cat2', self.RANGE_COL: 122, 'some_col': 'test2'},
        ]
        for x in rows:
            self.dynamo_client.put(x, self.table_name)

        keys_list_query = [
            {self.HASH_COL: 'cat1', self.RANGE_COL: 121},
            {self.HASH_COL: 'doesnt_exist', self.RANGE_COL: 40},
            {self.HASH_COL: 'cat2', self.RANGE_COL: 122},
        ]

        result = self.dynamo_client.batch_get_items_one_table(keys_list_query)

        self.assertEqual(len(result), 2)

        self.assertIn(rows[0], result)
        self.assertIn(rows[2], result)


    def test_delete(self):
        self.dynamo_client.put({self.HASH_COL: 'cat1', self.RANGE_COL: 123})
        self.dynamo_client.put({self.HASH_COL: 'cat2', self.RANGE_COL: 234})

        self.dynamo_client.delete(keys={self.HASH_COL: 'cat1', self.RANGE_COL: '123'})

        items = self.dynamo_client.get_by_scan()

        self.assertEqual(len(items), 1)
        self.assertEqual(items[0], {self.HASH_COL: 'cat2', self.RANGE_COL: 234})


    def test_get_table_keys(self):
        result1 = self.dynamo_client.get_table_keys()
        self.assertEqual(result1, ('hash_col', 'range_col'))

        result2 = self.dynamo_client.get_table_keys(table_name=self.table_name)
        self.assertEqual(result2, ('hash_col', 'range_col'))

    def test_get_table_indexes(self):
        indexes = self.dynamo_client.get_table_indexes()
        expected = {
            'autotest_index': {
                'projection_type': 'ALL',
                'hash_key':        'hash_col',
                'range_key':       'other_col',
                'provisioned_throughput': {
                    'write_capacity': 1,
                    'read_capacity':  1
                }
            }
        }
        self.assertDictEqual(expected, indexes)


    def test_batch_get_items_one_table(self):
        # If you want to stress test batch_get_items_one_table, use bigger numbers
        num_of_items = 5
        query_from = 2
        query_till = 4
        expected_items = query_till - query_from

        # Write items
        operations = []
        query_keys = []
        for i in range(num_of_items):
            item = {self.HASH_COL: f'cat{i%2}', self.RANGE_COL: i}
            operations.append({'Put': self.dynamo_client.build_put_query(item)})
            query_keys.append(item)
        for operations_chunk in chunks(operations, 10):
            self.dynamo_client.dynamo_client.transact_write_items(TransactItems=operations_chunk)
            time.sleep(1)  # cause the table has 10 write/sec capacity

        # Batch get items
        results = self.dynamo_client.batch_get_items_one_table(keys_list=query_keys[query_from:query_till])
        self.assertEqual(expected_items, len(results))


if __name__ == '__main__':
    unittest.main()



import boto3
import csv
import logging
import os
import unittest
from unittest.mock import patch, MagicMock

from sosw.components.config import ConfigSource


logging.getLogger('botocore').setLevel(logging.WARNING)

os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class FakeDynamo:

    def get_by_query(*args, **kwargs):
        pass  # This is patched later


class Config_UnitTestCase(unittest.TestCase):

    def setUp(self):
        with patch('sosw.components.config.DynamoConfig') as patch_dynamo:
            with patch('sosw.components.config.SSMConfig') as patch_ssm:
                self.config_source = ConfigSource(test=True)
                self.patched_ssm = patch_ssm
                self.patched_dynamo = patch_dynamo


    def test_get_config(self):
        # checks it calls dynamo config
        self.config_source.get_config('something')
        self.config_source.default_source.get_config.assert_called_once_with('something')


    def test_update_config(self):
        # checks it calls dynamo config
        self.config_source.update_config('name', 'value')
        self.config_source.default_source.update_config.assert_called_once_with('name', 'value')


    def test_default_sources(self):
        config_source = ConfigSource(test=True)

        self.assertFalse(hasattr(config_source, 'ssm_config'))
        self.assertTrue(hasattr(config_source, 'dynamo_config'))
        self.assertEqual(config_source.default_source, getattr(config_source, 'dynamo_config'))


    def test_custom_client(self):
        # checks it calls dynamo config
        config_source = ConfigSource(test=True, sources='SSM')

        self.assertTrue(hasattr(config_source, 'ssm_config'))
        self.assertFalse(hasattr(config_source, 'dynamo_config'))


    def test_custom_client__multiple(self):
        # checks it calls dynamo config
        config_source = ConfigSource(test=True, sources='SSM, Dynamo')

        self.assertTrue(hasattr(config_source, 'ssm_config'))
        self.assertTrue(hasattr(config_source, 'dynamo_config'))
        self.assertEqual(config_source.default_source, getattr(config_source, 'ssm_config'))


if __name__ == '__main__':
    unittest.main()



# import boto3
# import logging
# import time
# import os
#
#
# __author__ = "Nikolay Grishchenko"
# __email__ = "dev@bimpression.com"
# __version__ = "1.3"
# __license__ = "MIT"
# __status__ = "Production"
#
# __all__ = ['close_task']
#
# logger = logging.getLogger()
#
#
#
# def close_task(event, table=None, hash_key='task_id', range_key='labourer_id'):
#     """
#     Mark the task as completed in DynamoDB by updating the `completed_at` field with the current timestamp.
#     This function doesn't rely on components.dynamo_db because it can be used by the Worker for his own tasks
#     with his own configurations. And reinitialization of the client just for closing is too expensive.
#     Raw boto3 call here.
#
#     .. :note::
#        The current version asserts that Hash key is `String`.
#        Feel free to upgrade this for autodetection and make a PR.
#
#     :param dict event:   The Payload Worker Lambda has received. This method tries to extract key from it.
#
#     :param str table:       Name of the table to close task at. Default: `abs_tasks_running`
#     :param str hash_key:    Name of the Primary key (Hash key) in the table. Default: `task_id`
#     """
#
#     test = True if os.environ.get('STAGE') in ['test', 'autotest'] else False
#     table_name = table or ('autotest_sosw_tasks' if test else 'sosw_tasks')
#
#     hk = str(event.get(hash_key))
#     # rk = str(event.get(range_key, os.environ.get('AWS_LAMBDA_FUNCTION_NAME')))
#
#     client = boto3.client('dynamodb')
#
#     if hk:
#         response = client.update_item(TableName=table_name,
#                                       Key={hash_key: {'S': hk}},
#                                       UpdateExpression=" SET completed_at = :now",
#                                       ExpressionAttributeValues={':now': {'N': str(time.time())}})
#         logger.info(f"Marked task {hk} as completed")
#         logger.debug(f"DynamoDB response for marking task: {response}")
#
#     else:
#         logger.warning("No primary key for Tasks DB received. Someone other that adw_pull_controller invoked me.")
